from .transformation import Transformation
from .decimation import Decimation
from .fir_filtering import FIRFiltering
from .normalization import Normalization
from .window_sliding import WindowSliding
from .slicing import Slicing
from .diff_and_concat import DiffAndConcat
from .identity import Identity
from .custom import Custom
from .kalman_filter import KalmanFilter



import numpy as np

from .transformation import Transformation


class Slicing(Transformation):
    def __init__(self, amount: int, end: str) -> None:
        self.amount: int = amount
        self.end: str = end

    def __call__(self, x: np.ndarray) -> np.ndarray:
        if self.end == "left":
            return x[self.amount:]
        elif self.end == "right":
            return x[:-self.amount]
        else:
            raise ValueError("Unknown slicing end")



# regressors
from .regressors.approximate_gp_regressor import ApproximateGPRegressor
from .regressors.feedforward_regressor import FeedforwardRegressor
from .regressors.transformer_regressor import TransformerRegressor
from .regressors.transformer_fft_regressor import TransformerFFTRegressor
from .regressors.mamba_regressor import MambaRegressor
from .regressors.mamba_fft_regressor import MambaFFTRegressor

# classifiers
from .classifiers.feedforward_classifier import FeedforwardClassifier
from .classifiers.transformer_classifier import TransformerClassifier
from .classifiers.transformer_fft_classifier import TransformerFFTClassifier
from .classifiers.mamba_classifier import MambaClassifier
from .classifiers.mamba_fft_classifier import MambaFFTClassifier
from .classifiers.cnn_classifier import CNNClassifier



from torch import Tensor
from torch.nn import (
    Module,
    Linear,
    Sequential,
    ReLU,
    Sigmoid,
    Tanh,
    GELU,
    LayerNorm,
    Dropout,
    BatchNorm1d,
)

from einops import rearrange


def get_actvation_function(activation: str) -> Module:
    match activation:
        case 'relu':
            return ReLU()
        case 'sigmoid':
            return Sigmoid()
        case 'tanh':
            return Tanh()
        case 'gelu':
            return GELU()
        case _:
            raise ValueError(f'Activation function {activation} not supported')


class FeedforwardClassifier(Module):
    def __init__(self, *,
                 input_size: tuple[int, int, int],
                 dropout: float,
                 n_classes: int,
                 hidden_dims: list[int] = [],
                 use_layer_norm: bool = True,
                 use_batch_norm: bool = False,
                 activation: str = 'relu',
                 ) -> None:
        super().__init__()

        image_height, image_width, image_channels = input_size
        input_dim: int = image_height * image_width * image_channels

        layers: list[Module] = []

        # no hidden layers
        if not hidden_dims:
            self.net = Sequential(Linear(input_dim, n_classes))
            return

        # hidden layers
        layers.append(Linear(input_dim, hidden_dims[0]))
        n_layers = len(hidden_dims)

        for i in range(1, n_layers):
            # linear
            layers.append(Linear(hidden_dims[i-1], hidden_dims[i]))

            # batch norm
            if use_batch_norm:
                layers.append(BatchNorm1d(hidden_dims[i]))

            # layer norm
            if use_layer_norm:
                layers.append(LayerNorm(hidden_dims[i]))

            # activation
            layers.append(get_actvation_function(activation))

            # dropout
            layers.append(Dropout(dropout))

        layers.append(Linear(hidden_dims[-1], n_classes))

        self.net: Sequential = Sequential(*layers)

    def forward(self, x: Tensor) -> Tensor:
        x = rearrange(x, "b ... -> b (...)")

        x = self.net(x)

        return x



import platform
import os

import numpy as np

import torch
from torch import Tensor
from torch.nn import Identity
from torch.distributions import Distribution

from vclog import Logger

from ..dataclasses import PredictionInfo


def _get_device(device: str) -> torch.device:
    match device:
        case "cpu":
            return torch.device("cpu")
        case "cuda":
            return torch.device("cuda")
        case "auto":
            return torch.device("cuda" if torch.cuda.is_available() else "cpu")
        case _:
            raise ValueError("Invalid device")


def _get_dtype(dtype: str) -> torch.dtype:
    match dtype:
        case "float16":
            return torch.float16
        case "float32":
            return torch.float32
        case "float64":
            return torch.float64
        case _:
            raise ValueError("Invalid dtype")


class Archway:
    """A class for inference with a trained model.
    """

    def __init__(self, *,
                 path: str,
                 device: str = "auto",
                 dtype: str = "float32",
                 verbose: bool = True,
                 ) -> None:
        """Initialize the Archway object.

        Arguments
        ---------
        path : str
            The path to the directory containing the saved model. The directory should be named after the model.
        device : str, optional
            The device to perform inference on. Can be "cpu", "cuda", or "auto".
        dtype : str, optional
            The data type to use for inference. Can be "float16", "float32", or "float64".
        verbose : bool, optional
            Whether to log information about the inference process.

        Raises
        ------
        ValueError
            If the load path is a file path.
        """
        self.path: str = path if path.endswith("/") else f"{path}/"

        self.device: torch.device = _get_device(device)
        self.dtype: torch.dtype = _get_dtype(dtype)

        self.logger: Logger = Logger("catasta", disable=not verbose)

        self._load_model()
        self._log_inference_info()

    @torch.no_grad()
    def predict(self, input: np.ndarray | Tensor) -> PredictionInfo:
        """Perform inference with the model.

        Arguments
        ---------
        input : np.ndarray or Tensor
            The input data to make predictions on.

        Returns
        -------
        ~catasta.dataclasses.PredictionInfo
            The predicted values and standard deviations.
        """
        input_tensor: Tensor = torch.tensor(input) if isinstance(input, np.ndarray) else input
        input_tensor = input_tensor.to(self.device, self.dtype)

        output = self.likelihood(self.model(input_tensor))

        if isinstance(output, Distribution):
            predicted_output = output.mean.cpu().numpy()
            predicted_std = output.stddev.cpu().numpy()
        else:
            predicted_output = output.cpu().numpy()
            predicted_std = np.zeros_like(predicted_output)

        if predicted_output.ndim == 2:
            argmax: np.ndarray = np.argmax(predicted_output, axis=1)
        else:
            argmax: np.ndarray = np.array([])

        return PredictionInfo(
            value=predicted_output,
            std=predicted_std,
            argmax=argmax,
        )

    def _load_model(self) -> None:
        if "." in self.path:
            raise ValueError("load path must be a directory, not a file path")

        self.model = Identity()
        self.likelihood = Identity()
        for model in os.listdir(self.path):
            if os.path.basename(self.path.rstrip("/")).lower() == model.split(".")[0].lower():
                self.model = torch.load(os.path.join(self.path, model), map_location=self.device)
            elif "likelihood" in model.lower():
                self.likelihood = torch.load(os.path.join(self.path, model), map_location=self.device)

            self.logger.info(f"loaded {model} from {self.path}")

        if isinstance(self.model, Identity):
            raise ValueError("model not found")

        self.model.to(self.device, self.dtype)
        self.likelihood.to(self.device, self.dtype)

        self.model.eval()
        self.likelihood.eval()

    def _log_inference_info(self) -> None:
        n_params: int = sum(p.numel() for p in self.model.parameters() if p.requires_grad)
        device_name: str = torch.cuda.get_device_name() if self.device.type == "cuda" else platform.processor()
        self.logger.info(f"""inferring
    -> model: {self.model.__class__.__name__} ({n_params} trainable parameters)
    -> device: {self.device} ({device_name})
        """)



import os

from catasta.models import TransformerRegressor, ApproximateGPRegressor
from catasta.datasets import RegressionDataset
from catasta.scaffolds import RegressionScaffold
from catasta.dataclasses import RegressionEvalInfo
from catasta.transformations import (
    Normalization,
    Decimation,
    WindowSliding,
    Slicing,
    Custom,
)

from vclog import Logger


def vanilla() -> None:
    n_dim: int = 768
    dataset_root: str = "tests/data/nylon_elastic/strain/"
    input_trasnsformations = [
        Custom(lambda x: x[500_000:1_000_000]),
        Normalization("minmax"),
        Decimation(decimation_factor=100),
        WindowSliding(window_size=n_dim, stride=1),
    ]
    output_trasnsformations = [
        Custom(lambda x: x[500_000:1_000_000]),
        Normalization("minmax"),
        Decimation(decimation_factor=100),
        Slicing(amount=n_dim - 1, end="left"),
    ]

    n_files = len(os.listdir(dataset_root))
    train_split = (n_files-1) / n_files
    dataset = RegressionDataset(
        root=dataset_root,
        input_transformations=input_trasnsformations,
        output_transformations=output_trasnsformations,
        splits=(train_split, 1-train_split, 0.0)
    )

    model = TransformerRegressor(
        context_length=n_dim,
        n_patches=8,
        d_model=8,
        n_heads=4,
        n_layers=2,
        feedforward_dim=4,
        head_dim=4,
        dropout=0.0,
    )
    scaffold = RegressionScaffold(
        model=model,
        dataset=dataset,
        optimizer="adamw",
        loss_function="mse",
    )

    scaffold.train(
        epochs=100,
        batch_size=256,
        lr=1e-4,
    )
    info: RegressionEvalInfo = scaffold.evaluate()
    Logger.debug(info)

    scaffold.save(path="tests/models/", to_onnx=True)


def gp():
    n_dim: int = 768
    dataset_root: str = "tests/data/nylon_elastic/strain/"
    input_trasnsformations = [
        Custom(lambda x: x[500_000:1_000_000]),
        Normalization("minmax"),
        Decimation(decimation_factor=100),
        WindowSliding(window_size=n_dim, stride=1),
    ]
    output_trasnsformations = [
        Custom(lambda x: x[500_000:1_000_000]),
        Normalization("minmax"),
        Decimation(decimation_factor=100),
        Slicing(amount=n_dim - 1, end="left"),
    ]

    n_files = len(os.listdir(dataset_root))
    train_split = (n_files-1) / n_files
    dataset = RegressionDataset(
        root=dataset_root,
        input_transformations=input_trasnsformations,
        output_transformations=output_trasnsformations,
        splits=(train_split, 0.0, 1-train_split)
    )

    model = ApproximateGPRegressor(
        context_length=n_dim,
        n_inducing_points=128,
    )
    scaffold = RegressionScaffold(
        model=model,
        dataset=dataset,
        optimizer="adamw",
        loss_function="variational_elbo",
    )

    scaffold.train(
        epochs=100,
        batch_size=256,
        lr=1e-4,
    )
    info: RegressionEvalInfo = scaffold.evaluate()
    Logger.debug(info)

    scaffold.save(path="tests/models/gp/")


if __name__ == '__main__':
    # vanilla()
    gp()



from catasta import Scaffold, CatastaDataset
from catasta.models import TransformerRegressor
from catasta.transformations import (
    Normalization,
    WindowSliding,
    Slicing,
    Custom,
)

from vclog import Logger


def main() -> None:
    n_dim: int = 768
    dataset_root: str = "data/nylon_wire/"
    input_trasnsformations = [
        Custom(lambda x: x[:10_000]),
        Normalization("minmax"),
        WindowSliding(window_size=n_dim, stride=1),
    ]
    output_trasnsformations = [
        Custom(lambda x: x[:10_000]),
        Normalization("minmax"),
        Slicing(amount=n_dim - 1, end="left"),
    ]

    dataset = CatastaDataset(
        root=dataset_root,
        task="regression",
        input_transformations=input_trasnsformations,
        output_transformations=output_trasnsformations,
    )

    model = TransformerRegressor(
        context_length=n_dim,
        n_patches=8,
        d_model=8,
        n_heads=4,
        n_layers=2,
        feedforward_dim=4,
        head_dim=4,
        dropout=0.0,
    )
    scaffold = Scaffold(
        model=model,
        dataset=dataset,
        optimizer="adamw",
        loss_function="mse",
    )

    scaffold.train(
        epochs=100,
        batch_size=256,
        lr=1e-3,
        final_lr=1e-4,
        early_stopping=True,
    )
    info = scaffold.evaluate()

    Logger.debug(info)


if __name__ == '__main__':
    main()



import numpy as np
import matplotlib.pyplot as plt

from catasta.models import ApproximateGPRegressor
from catasta.datasets import RegressionDataset
from catasta.dataclasses import RegressionEvalInfo, RegressionTrainInfo, RegressionPrediction
from catasta.utils import get_optimizer, get_objective_function

from vclog import Logger

import torch
from torch import Tensor
from torch.nn import Module
from torch.utils.data import DataLoader
from torch.optim import Optimizer
from torch.distributions import Distribution

from gpytorch.distributions import MultivariateNormal
from gpytorch.mlls import MarginalLogLikelihood
from gpytorch.likelihoods import GaussianLikelihood
from torch import nn
from torch.fft import fft
from torch.nn import Module, Sequential, LayerNorm, Linear, GELU, Softmax

from einops import rearrange
from einops.layers.torch import Rearrange

from torch.nn import Module, Linear, Sequential, ReLU, BatchNorm1d, Dropout


class FeedforwardRegressor(Module):
    def __init__(self, *,
                 input_dim: int,
                 hidden_dims: list[int],
                 output_dim: int,
                 dropout: float,
                 ) -> None:
        super().__init__()

        n_layers: int = len(hidden_dims)
        layers: list[Module] = []
        for i in range(n_layers):
            if i == 0:
                layers.append(Linear(input_dim, hidden_dims[i]))
            else:
                layers.append(Linear(hidden_dims[i-1], hidden_dims[i]))

            layers.append(BatchNorm1d(hidden_dims[i]))
            layers.append(ReLU())
            layers.append(Dropout(dropout))

        layers.append(Linear(hidden_dims[-1], output_dim))

        self.model: Sequential = Sequential(*layers)

    def forward(self, x: Tensor) -> Tensor:
        x = self.model(x)

        return x


def posemb_sincos_1d(patches: Tensor, temperature: int = 10000) -> Tensor:
    n: int = patches.shape[1]
    d_model: int = patches.shape[2]
    device: torch.device = patches.device
    dtype: torch.dtype = patches.dtype

    if (d_model % 2) != 0:
        raise ValueError(f'feature dimension must be multiple of 2 for sincos emb. got {d_model}')

    n_tensor: Tensor = torch.arange(n, device=device)
    omega: Tensor = torch.arange(d_model // 2, device=device) / (d_model // 2 - 1)
    omega = 1.0 / (temperature ** omega)

    n_tensor = n_tensor.flatten()[:, None] * omega[None, :]
    pe: Tensor = torch.cat((n_tensor.sin(), n_tensor.cos()), dim=1)

    return pe.to(dtype)


class FeedForward(Module):
    def __init__(self, *,
                 d_model: int,
                 hidden_dim: int,
                 ) -> None:
        super().__init__()

        self.net = Sequential(
            LayerNorm(d_model),
            Linear(d_model, hidden_dim),
            GELU(),
            Linear(hidden_dim, d_model),
        )

    def forward(self, x: Tensor) -> Tensor:
        return self.net(x)


class Attention(nn.Module):
    def __init__(self, *,
                 d_model: int,
                 n_heads: int = 8,
                 head_dim: int = 64,
                 ) -> None:
        super().__init__()

        inner_dim: int = head_dim * n_heads
        self.n_heads: int = n_heads
        self.scale: int = head_dim ** -0.5
        self.norm = LayerNorm(d_model)

        self.attend = Softmax(dim=-1)

        self.to_qkv = Linear(d_model, inner_dim * 3, bias=False)
        self.to_out = Linear(inner_dim, d_model, bias=False)

    def forward(self, x: Tensor) -> Tensor:
        x = self.norm(x)

        qkv: Tensor = self.to_qkv(x).chunk(3, dim=-1)
        q, k, v = map(lambda t: rearrange(t, 'b n (h d) -> b h n d', h=self.n_heads), qkv)

        dots: Tensor = torch.matmul(q, k.transpose(-1, -2)) * self.scale

        attn: Tensor = self.attend(dots)

        out: Tensor = torch.matmul(attn, v)
        out = rearrange(out, 'b h n d -> b n (h d)')

        return self.to_out(out)


class Transformer(nn.Module):
    def __init__(self, *,
                 d_model: int,
                 n_layers: int,
                 n_heads: int,
                 head_dim: int,
                 feedforward_dim: int,
                 ) -> None:
        super().__init__()

        self.norm = nn.LayerNorm(d_model)
        self.layers = nn.ModuleList([])
        for _ in range(n_layers):
            self.layers.append(nn.ModuleList([
                Attention(d_model=d_model, n_heads=n_heads, head_dim=head_dim),
                FeedForward(d_model=d_model, hidden_dim=feedforward_dim)
            ]))

    def forward(self, x: Tensor) -> Tensor:
        for attn, ff in self.layers:  # type: ignore
            x = attn(x) + x
            x = ff(x) + x
        return self.norm(x)


class FFTTransformerRegressor(Module):
    def __init__(self, *,
                 context_length: int,
                 n_patches: int,
                 output_dim: int,
                 d_model: int,
                 n_layers: int,
                 n_heads: int,
                 feedforward_dim: int,
                 head_dim: int,
                 ) -> None:
        super().__init__()
        patch_dim: int = context_length // n_patches
        freq_patch_dim: int = patch_dim * 2

        if context_length % patch_dim != 0:
            raise ValueError(f"sequence length {context_length} must be divisible by patch size {patch_dim}")

        self.to_patch_embedding = nn.Sequential(
            Rearrange('b c (n p) -> b n (p c)', p=patch_dim),
            LayerNorm(patch_dim),
            Linear(patch_dim, d_model),
            LayerNorm(d_model),
        )

        self.to_freq_embedding = nn.Sequential(
            Rearrange('b c (n p) ri -> b n (p ri c)', p=patch_dim),
            LayerNorm(freq_patch_dim),
            Linear(freq_patch_dim, d_model),
            LayerNorm(d_model),
        )

        self.transformer = Transformer(
            d_model=d_model,
            n_layers=n_layers,
            n_heads=n_heads,
            head_dim=head_dim,
            feedforward_dim=feedforward_dim,
        )

        self.linear_head = nn.Linear(d_model, output_dim)

    def forward(self, input: Tensor) -> Tensor:
        input = rearrange(input, 'b s -> b 1 s')
        freqs: Tensor = torch.view_as_real(fft(input))

        x: Tensor = self.to_patch_embedding(input)
        f: Tensor = self.to_freq_embedding(freqs)

        x += posemb_sincos_1d(x)
        f += posemb_sincos_1d(f)

        x = torch.cat((x, f), dim=1)
        x = self.transformer(x)
        x = x.mean(dim=1)

        return x


class GaussianRegressionScaffold:
    def __init__(self, *,
                 pre_model: Module,
                 model: Module,
                 dataset: RegressionDataset,
                 optimizer: str = "adam",
                 loss_function: str = "variational_elbo",
                 ) -> None:
        self.device: torch.device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
        self.dtype: torch.dtype = torch.float32

        self.pre_model: Module = pre_model.to(self.device)

        self.model: Module = model.to(self.device)
        self.likelihood = GaussianLikelihood().to(self.device)
        self.dataset: RegressionDataset = dataset

        self.optimizer_id: str = optimizer
        self.loss_function_id: str = loss_function

        self.logger: Logger = Logger("catasta")
        self.logger.info(f"using device: {self.device}")

    def train(self,
              epochs: int = 100,
              batch_size: int = 32,
              lr: float = 1e-3,
              ) -> RegressionTrainInfo:
        self.pre_model.train()
        self.model.train()
        self.likelihood.train()

        data_loader: DataLoader = DataLoader(self.dataset.train, batch_size=batch_size, shuffle=False)

        optimizer: Optimizer | None = get_optimizer(self.optimizer_id, [self.model, self.likelihood, self.pre_model], lr)
        if optimizer is None:
            raise ValueError(f"invalid optimizer id: {self.optimizer_id}")

        mll: MarginalLogLikelihood | None = get_objective_function(self.loss_function_id, self.model, self.likelihood, len(self.dataset.train))
        if mll is None:
            raise ValueError(f"invalid loss function id: {self.loss_function_id}")

        losses: list[float] = []
        for i in range(epochs):
            batch_losses: list[float] = []
            for j, (x_batch, y_batch) in enumerate(data_loader):
                optimizer.zero_grad()

                x_batch = x_batch.to(self.device, dtype=self.dtype)
                y_batch = y_batch.to(self.device, dtype=self.dtype)

                x_batch = self.pre_model(x_batch)
                output: MultivariateNormal = self.model(x_batch)

                loss: Tensor = -mll(output, y_batch)  # type: ignore
                batch_losses.append(loss.item())
                loss.backward()

                optimizer.step()

                self.logger.info(f"epoch {i}/{epochs} | {int((i/epochs)*100+(j/len(data_loader))*100/epochs)}% | loss: {loss.item():.4f}", flush=True)

            losses.append(np.mean(batch_losses).astype(float))

        self.logger.info(f'epoch {epochs}/{epochs} | 100% | loss: {np.min(losses):.4f}')

        return RegressionTrainInfo(np.array(losses))

    @torch.no_grad()
    def predict(self, input: np.ndarray | Tensor) -> RegressionPrediction:
        self.model.eval()
        self.likelihood.eval()
        self.pre_model.eval()

        input_tensor: Tensor = torch.tensor(input) if isinstance(input, np.ndarray) else input
        input_tensor = input_tensor.to(self.device, dtype=self.dtype)

        input_tensor = self.pre_model(input_tensor)

        output: Distribution = self.likelihood(self.model(input_tensor))

        mean: np.ndarray = output.mean.cpu().numpy()
        std: np.ndarray = output.stddev.cpu().numpy()

        return RegressionPrediction(mean, std)

    @torch.no_grad()
    def evaluate(self) -> RegressionEvalInfo:
        test_index: int = np.floor(len(self.dataset) * self.dataset.splits[0]).astype(int)
        test_x: np.ndarray = self.dataset.inputs[test_index:]
        test_y: np.ndarray = self.dataset.outputs[test_index:].flatten()

        if self.dataset.test is None:
            raise ValueError(f"test split must be greater than 0")

        data_loader = DataLoader(self.dataset.test, batch_size=1, shuffle=False)

        means: np.ndarray = np.array([])
        stds: np.ndarray = np.array([])
        for x_batch, _ in data_loader:
            output: RegressionPrediction = self.predict(x_batch)
            means = np.concatenate([means, output.prediction])
            if output.stds is not None:
                stds = np.concatenate([stds, output.stds])

        if len(means) != len(test_y):
            min_len: int = min(len(means), len(test_y))
            means = means[-min_len:]
            stds = stds[-min_len:]
            test_y = test_y[-min_len:]

        return RegressionEvalInfo(test_x, test_y, means, stds)


def main() -> None:
    # dataset_root: str = "tests/data/wire_lisbeth/strain/"
    dataset_root: str = "tests/data/nylon_carmen/paper/strain/mixed_10_20/"
    # dataset_root: str = "tests/data/nylon_carmen_elasticband/paper/strain/mixed_10_20/"

    context_length: int = 1024
    d_model: int = 32

    # 512, 4, 1, 16, 2, 2, 32, 4
    # dataset = RegressionDataset(
    #     root=dataset_root,
    #     context_length=context_length,
    #     splits=(7/8, 1/8, 0.0),
    # )
    pre_model = FFTTransformerRegressor(
        context_length=context_length,
        n_patches=4,
        output_dim=1,
        d_model=d_model,
        n_heads=2,
        n_layers=2,
        feedforward_dim=32,
        head_dim=16,
    )
    # pre_model = FeedforwardRegressor(
    #     input_dim=context_length,
    #     hidden_dims=[128, 64, 32],
    #     output_dim=d_model,
    #     dropout=0.2,
    # )
    n_inducing_points: int = 128
    dataset = RegressionDataset(
        root=dataset_root,
        context_length=context_length,
        prediction_length=1,
        splits=(7/8, 0.0, 1/8),
    )
    gp_model = ApproximateGPRegressor(
        n_inducing_points=n_inducing_points,
        n_inputs=d_model,
        kernel="rff",
        mean="zero"
    )
    scaffold = GaussianRegressionScaffold(
        model=gp_model,
        pre_model=pre_model,
        dataset=dataset,
        optimizer="adamw",
        loss_function="variational_marginal_log",
    )

    train_info: RegressionTrainInfo = scaffold.train(
        epochs=1000,
        batch_size=128,
        lr=1e-3,
    )

    Logger.debug(f"min train loss: {np.min(train_info.train_loss):.4f}")

    plt.figure(figsize=(30, 20))
    plt.plot(train_info.train_loss, label="train loss", color="black")
    plt.legend()
    plt.show()

    info: RegressionEvalInfo = scaffold.evaluate()

    plt.figure(figsize=(30, 20))
    plt.plot(info.predicted, label="predictions", color="red")
    plt.plot(info.real, label="real", color="black")
    plt.fill_between(range(len(info.predicted)), info.predicted-1*info.stds, info.predicted+1*info.stds, color="red", alpha=0.2)  # type: ignore
    plt.legend()
    plt.show()

    Logger.debug(info)


if __name__ == '__main__':
    main()



from torch.nn import Module, Identity

from gpytorch.likelihoods import (
    GaussianLikelihood,
    BernoulliLikelihood,
    LaplaceLikelihood,
    SoftmaxLikelihood,
    StudentTLikelihood,
    BetaLikelihood,
)


id_to_likelihood = {
    "gaussian": GaussianLikelihood,
    "bernoulli": BernoulliLikelihood,
    "laplace": LaplaceLikelihood,
    "softmax": SoftmaxLikelihood,
    "studentt": StudentTLikelihood,
    "beta": BetaLikelihood,
}


def get_likelihood(id: str | Module | None) -> Module:
    if id is None:
        return Identity()
    if isinstance(id, Module):
        return id

    likelihood = id_to_likelihood.get(id.lower(), None)
    if likelihood is None:
        raise ValueError(f"Likelihood {id} not recognized")

    return likelihood()



from torch import Tensor
from torch.nn import (
    Module,
    Conv2d,
    MaxPool2d,
    Linear,
    Flatten,
    Dropout,
    Sequential,
    ReLU,
    Sigmoid,
    Tanh,
    GELU,
)

from einops import rearrange


def get_actvation_function(activation: str) -> Module:
    match activation:
        case 'relu':
            return ReLU()
        case 'sigmoid':
            return Sigmoid()
        case 'tanh':
            return Tanh()
        case 'gelu':
            return GELU()
        case _:
            raise ValueError(f'Activation function {activation} not supported')


class CNNClassifier(Module):
    def __init__(self, *,
                 input_shape: tuple[int, int, int],
                 n_classes: int,
                 conv_out_channels: list[int],
                 conv_kernel_sizes: list[int],
                 conv_strides: list[int],
                 conv_paddings: list[int],
                 pooling_kernel_sizes: list[int],
                 pooling_strides: list[int],
                 pooling_paddings: list[int],
                 feedforward_dims: list[int],
                 dropout: float = 0.5,
                 activation: str = 'relu',
                 ) -> None:
        super().__init__()

        n_input_channels: int = input_shape[-1]

        conv_layers: list[Module] = []
        for i in range(len(conv_out_channels)):
            conv2d = Conv2d(
                in_channels=n_input_channels if i == 0 else conv_out_channels[i - 1],
                out_channels=conv_out_channels[i],
                kernel_size=conv_kernel_sizes[i],
                stride=conv_strides[i],
                padding=conv_paddings[i],
            )

            activation_function: Module = get_actvation_function(activation)

            max_pool2d = MaxPool2d(
                kernel_size=pooling_kernel_sizes[i],
                stride=pooling_strides[i],
                padding=pooling_paddings[i],
            )

            conv_layers.append(conv2d)
            conv_layers.append(activation_function)
            conv_layers.append(max_pool2d)

        flatten = Flatten()

        prev_neurons: int = conv_out_channels[-1] * (input_shape[0] // 2 ** len(conv_out_channels)) * (input_shape[1] // 2 ** len(conv_out_channels))
        dense_layers: list[Module] = []
        for i in range(len(feedforward_dims)):
            linear = Linear(
                in_features=prev_neurons if i == 0 else feedforward_dims[i - 1],
                out_features=feedforward_dims[i],
            )

            activation_function: Module = get_actvation_function(activation)
            dropout_layer = Dropout(p=dropout)

            dense_layers.append(linear)
            dense_layers.append(activation_function)
            dense_layers.append(dropout_layer)

        linear = Linear(
            in_features=feedforward_dims[-1],
            out_features=n_classes,
        )

        dense_layers.append(linear)

        self.net: Sequential = Sequential(*conv_layers, flatten, *dense_layers)

    def forward(self, x: Tensor) -> Tensor:
        x = rearrange(x, 'b h w c -> b c h w')

        return self.net(x)



import os

import numpy as np
from PIL import Image

from catasta import Scaffold, CatastaDataset, Archway
from catasta.models import FeedforwardClassifier
from catasta.dataclasses import ClassificationEvalInfo


def main() -> None:
    model = FeedforwardClassifier(
        input_size=(28, 28, 3),
        n_classes=10,
        hidden_dims=[16, 16, 16],
        activation="relu",
        dropout=0.1,
    )

    dataset = CatastaDataset("data/reduced_mnist", task="classification")

    scaffold = Scaffold(
        model=model,
        dataset=dataset,
        optimizer="adamw",
        loss_function="cross_entropy",
    )

    scaffold.train(
        epochs=100,
        batch_size=128,
        lr=1e-3,
    )

    eval_info = scaffold.evaluate()
    print(eval_info)

    save_model_path = "saved_models/"
    scaffold.save(save_model_path)

    archway = Archway(
        path=save_model_path,
    )

    imgs = []
    dataset_root = "data/reduced_mnist/val/0"
    for file in os.listdir(dataset_root):
        img = Image.open(os.path.join(dataset_root, file))
        img = np.array(img.convert("RGB"))
        imgs.append(img)

    input = np.array(imgs)
    prediction = archway.predict(input)

    predicted_output = prediction.argmax
    true_labels = np.zeros(len(predicted_output)).astype(int)

    info = ClassificationEvalInfo(
        true_labels=true_labels,
        predicted_labels=predicted_output,
        n_classes=10,
    )

    print(info)


if __name__ == '__main__':
    main()



import numpy as np


class ClassificationEvalInfo:
    """A class for evaluating classification models.

    Attributes
    ----------
    true_labels : np.ndarray
        The true labels of the data.
    predicted_labels : np.ndarray
        The predicted labels of the data.
    n_classes : int
        The number of classes in the data.
    confusion_matrix : np.ndarray
        The confusion matrix of the data.
    tp : np.ndarray
        The true positives of the data.
    fp : np.ndarray
        The false positives of the data.
    fn : np.ndarray
        The false negatives of the data.
    tn : np.ndarray
        The true negatives of the data.
    accuracy : np.ndarray
        The accuracy of the data.
    precision : np.ndarray
        The precision of the data.
    sensitivity : np.ndarray
        The sensitivity of the data.
    specificity : np.ndarray
        The specificity of the data.
    f1 : np.ndarray
        The f1 score of the data.
    """

    def __init__(self, *,
                 true_labels: np.ndarray,
                 predicted_labels: np.ndarray,
                 n_classes: int | None = None,
                 ) -> None:
        """Initialize the ClassificationEvalInfo object.

        Arguments
        ---------
        true_labels : np.ndarray
            The true labels of the data.
        predicted_labels : np.ndarray
            The predicted labels of the data.
        n_classes : int, optional
            The number of classes in the data. If not provided, it will be inferred from the true labels.
        """
        self.true_labels: np.ndarray = true_labels
        self.predicted_labels: np.ndarray = predicted_labels
        self.n_classes: int | None = n_classes

        self.confusion_matrix: np.ndarray = self._compute_confusion_matrix()
        self.tp: np.ndarray = np.diag(self.confusion_matrix)
        self.fp: np.ndarray = np.sum(self.confusion_matrix, axis=0) - self.tp
        self.fn: np.ndarray = np.sum(self.confusion_matrix, axis=1) - self.tp
        self.tn: np.ndarray = np.sum(self.confusion_matrix) - (self.tp + self.fp + self.fn)

        self.accuracy: np.ndarray = (self.tp + self.tn) / (self.tp + self.fp + self.fn + self.tn)
        self.precision: np.ndarray = self.tp / (self.tp + self.fp)
        self.sensitivity: np.ndarray = self.tp / (self.tp + self.fn)
        self.specificity: np.ndarray = self.tn / (self.tn + self.fp)
        self.f1: np.ndarray = 2 * (self.precision * self.sensitivity) / (self.precision + self.sensitivity)

    def _compute_confusion_matrix(self) -> np.ndarray:
        classes: int = len(np.unique(self.true_labels)) if self.n_classes is None else self.n_classes

        matrix: np.ndarray = np.zeros((classes, classes), dtype=int)

        for true, pred in zip(self.true_labels, self.predicted_labels):
            matrix[true, pred] += 1

        return matrix

    def __repr__(self) -> str:
        accuracy: str = f"accuracy:{self.accuracy.mean(): .4f}"
        precision: str = f"precision:{self.precision.mean(): .4f}"
        sensitivity: str = f"sensitivity:{self.sensitivity.mean(): .4f}"
        specificity: str = f"specificity:{self.specificity.mean(): .4f}"
        f1: str = f"f1:{self.f1.mean(): .4f}"
        confusion_matrix: str = f"confusion_matrix:\n{self.confusion_matrix}"

        return "\n".join([accuracy, precision, sensitivity, specificity, f1, confusion_matrix])



from .optimizer_factory import get_optimizer
from .loss_function_factory import get_loss_function, available_loss_functions
from .objective_function_factory import get_objective_function
from .training_logger import TrainingLogger
from .model_state_manager import ModelStateManager
from .likelihood_factory import get_likelihood



import numpy as np

from catasta.transformations import WindowSliding, Slicing


def main() -> None:
    x: np.ndarray = np.arange(0, 100)
    window_size: int = 10
    stride: int = 1
    window_sliding: WindowSliding = WindowSliding(
        window_size=window_size, stride=stride
    )
    y: np.ndarray = window_sliding(x)
    slicing: Slicing = Slicing(amount=window_size - 1, end="left")

    x = slicing(x)

    for x, y in zip(x, y):
        print(y, x)


if __name__ == "__main__":
    main()



import os

import numpy as np
import pandas as pd


noise_mean: float = 0.0
noise_std: float = 0.02
n_samples: int = 1000

true_input: np.ndarray = np.linspace(-0.5, 1.5, n_samples)
input = np.hstack([np.linspace(-0.2, 0.2, n_samples//2), np.linspace(0.6, 1, n_samples//2)])


def f(x: np.ndarray, noise: np.ndarray) -> np.ndarray:
    return x + 0.3 * np.sin(2 * np.pi * (x + noise)) + 0.2 * np.sin(4 * np.pi * (x + noise)) + noise


output = f(input, np.random.normal(noise_mean, noise_std, n_samples))
true_output = f(true_input, np.zeros_like(true_input))

# divide into train and val
n_train = 800
n_val = n_samples - n_train

# take random samples
train_indices = np.random.choice(n_samples, n_train, replace=False)
val_indices = np.setdiff1d(np.arange(n_samples), train_indices)

train_true_input = true_input[train_indices]
train_true_output = true_output[train_indices]
train_input = input[train_indices]
train_output = output[train_indices]

val_true_input = true_input[val_indices]
val_true_output = true_output[val_indices]
val_input = input[val_indices]
val_output = output[val_indices]

data = {
    "input": input,
    "output": output,
    "true_input": true_input,
    "true_output": true_output,
}

pd.DataFrame(data).to_csv("data.csv", index=False)

train_data = {
    "input": train_input,
    "output": train_output,
    "true_input": train_true_input,
    "true_output": train_true_output,
}

if not os.path.exists("train"):
    os.makedirs("train")

pd.DataFrame(train_data).to_csv("train/data.csv", index=False)

val_data = {
    "input": val_input,
    "output": val_output,
    "true_input": val_true_input,
    "true_output": val_true_output,
}

if not os.path.exists("val"):
    os.makedirs("val")

pd.DataFrame(val_data).to_csv("val/data.csv", index=False)



import numpy as np


class RegressionEvalInfo:
    """A class for evaluating regression models.

    Attributes
    ----------
    true_input : np.ndarray
        The true input of the data.
    true_output : np.ndarray
        The true output of the data.
    predicted_output : np.ndarray
        The predicted output of the data.
    predicted_std : np.ndarray
        The predicted standard deviation of the data.
    mae : float
        The mean absolute error of the data.
    mse : float
        The mean squared error of the data.
    rmse : float
        The root mean squared error of the data.
    r2 : float
        The R2 score of the data.
    mape : float
        The mean absolute percentage error of the data.
    smape : float
        The symmetric mean absolute percentage error of the data.
    mase : float
        The mean absolute scaled error of the data.
    masep : float
        The mean absolute scaled error percentage of the data.
    """

    def __init__(self, true_input: np.ndarray, true_output: np.ndarray, predicted_output: np.ndarray, predicted_std: np.ndarray) -> None:
        """Initialize the RegressionEvalInfo object. 

        Arguments
        ---------
        true_input : np.ndarray
            The true input of the data.
        true_output : np.ndarray
            The true output of the data.
        predicted_output : np.ndarray
            The predicted output of the data.
        predicted_std : np.ndarray
            The predicted standard deviation of the data.
        """
        self.true_input: np.ndarray = true_input
        self.predicted_output: np.ndarray = predicted_output
        self.true_output: np.ndarray = true_output
        self.predicted_std: np.ndarray = predicted_std

        self.mae: float = np.mean(np.abs(predicted_output - true_output)).astype(float)
        self.mse: float = np.mean((predicted_output - true_output)**2).astype(float)
        self.rmse: float = np.sqrt(np.mean((predicted_output - true_output)**2))
        self.r2: float = 1 - np.sum((predicted_output - true_output)**2) / np.sum((true_output - np.mean(true_output))**2).astype(float)
        self.mape: float = np.mean(np.abs((predicted_output - true_output) / true_output)).astype(float)
        self.smape: float = np.mean(np.abs(predicted_output - true_output) / (np.abs(predicted_output) + np.abs(true_output))).astype(float)
        self.mase: float = np.mean(np.abs(predicted_output - true_output) / np.mean(np.abs(true_output))).astype(float)
        self.masep: float = np.mean(np.abs(predicted_output - true_output) / np.mean(np.abs(true_output - np.mean(true_output)))).astype(float)

    def __repr__(self) -> str:
        mae_str: str = f"MAE: {self.mae:.4f}\n"
        mse_str: str = f"MSE: {self.mse:.4f}\n"
        rmse_str: str = f"RMSE: {self.rmse:.4f}\n"
        r2_str: str = f"R2: {self.r2:.4f}\n"
        mape_str: str = f"MAPE: {self.mape:.4f}\n"
        smape_str: str = f"SMAPE: {self.smape:.4f}\n"
        mase_str: str = f"MASE: {self.mase:.4f}\n"
        masep_str: str = f"MASEP: {self.masep:.4f}\n"

        return mae_str + mse_str + rmse_str + r2_str + mape_str + smape_str + mase_str + masep_str



from typing import NamedTuple
import numpy as np


class TrainInfo(NamedTuple):
    """NamedTuple for storing training information.

    Attributes
    ----------
    train_losses : np.ndarray
        Array of the training losses for each epoch.
    best_train_loss : float
        The minimum value of the training losses.
    train_accuracies : np.ndarray
        Array of the training accuracies for each epoch. If the task is regression, this array will be empty.
    best_train_accuracy : float
        The maximum value of the training accuracies. If the task is regression, this value will be -inf.
    val_losses : np.ndarray
        Array of the validation losses for each epoch.
    best_val_loss : float
        The minimum value of the validation losses.
    val_accuracies : np.ndarray
        Array of the validation accuracies for each epoch. If the task is regression, this array will be empty.
    best_val_accuracy : float
        The maximum value of the validation accuracies. If the task is regression, this value will be -inf.
    lr_values : np.ndarray
        Array of the learning rates for each epoch.
    """
    train_losses: np.ndarray
    best_train_loss: float
    train_accuracies: np.ndarray
    best_train_accuracy: float

    val_losses: np.ndarray
    best_val_loss: float
    val_accuracies: np.ndarray
    best_val_accuracy: float

    lr_values: np.ndarray



import os
from typing import NamedTuple, Literal

import pandas as pd
import numpy as np
from PIL import Image

import torch
from torch import Tensor
from torch.utils.data import Dataset

from vclog import Logger

from ..transformations import Transformation


def scan_classes(dir: str) -> list[str]:
    classes: list[str] = []
    for entry in os.scandir(dir):
        if entry.is_dir():
            classes.append(entry.name)

    classes.sort()

    return classes


def scan_splits(root: str) -> tuple[str, str, str]:
    train_dir_name: str = ""
    validation_dir_name: str = ""
    test_dir_name: str = ""

    splits: list[str] = scan_classes(root)
    for split in splits:
        if "train" in split:
            train_dir_name = split
        elif "val" in split:
            validation_dir_name = split
        elif "test" in split:
            test_dir_name = split

    if not train_dir_name:
        raise ValueError(f"a directory that contains the subword 'train' must be present in {root}")
    elif not validation_dir_name and not test_dir_name:
        raise ValueError(f"at least a validation or test split must be present in {root}")
    elif not validation_dir_name and test_dir_name:
        Logger("catasta").warning(f"no validation split found in {root}. Using test split as validation split.")
        validation_dir_name = test_dir_name
    elif not test_dir_name and validation_dir_name:
        Logger("catasta").warning(f"no test split found in {root}. Using validation split as test split.")
        test_dir_name = validation_dir_name

    return train_dir_name, validation_dir_name, test_dir_name


class CatastaDataset:
    """Class to handle Catasta datasets. Catasta datasets are organized in a directory with the following structure:
    root
    ├── train
    ├── val
    └── test
    Each of the train, val, and test directories contain CSV or image files. For regression tasks, the CSV files must have the columns 'input' and 'output'.

    Attributes
    ----------
    root : str
        The root directory of the Catasta dataset.
    task : Literal["regression", "classification"]
        The task of the dataset.
    train : Dataset
        The training subset of the dataset.
    validation : Dataset
        The validation subset of the dataset.
    test : Dataset
        The test subset of the dataset.
    """

    def __init__(self,
                 root: str,
                 task: Literal["regression", "classification"],
                 input_transformations: list[Transformation] = [],
                 output_transformations: list[Transformation] = [],
                 input_name: str = "input",
                 output_name: str = "output",
                 ) -> None:
        """Initialize the CatastaDataset object.

        Parameters
        ----------
        root : str_
            The root directory of the Catasta dataset.
        task : Literal["regression", "classification"]
            The task of the dataset.
        input_transformations : list[~catasta.transformations.Transformation], optional
            A list of transformations to apply to the input data, by default [].
        output_transformations : list[~catasta.transformations.Transformation], optional
            A list of transformations to apply to the output data, by default [].
        input_name : str, optional
            The name of the column in the CSV files that contains the input data, by default "input".
        output_name : str, optional
            The name of the column in the CSV files that contains the output data, by default "output".

        Raises
        ------
        ValueError
            If the root directory does not exist.
            If the task is not 'regression' or 'classification'.
            If the train split is not found.
            If the validation split is not found and the test split is not found.
            If the directories do not contain at least one CSV file (regression).
            If the number of classes in the train and validation splits are not the same (classification).
            If the number of classes in the train and test splits are not the same (classification).
        """
        self.task: Literal["regression", "classification"] = task
        self.root: str = root if root.endswith("/") else root + "/"
        if not os.path.isdir(self.root):
            raise ValueError(f"root must be a directory. Found {self.root}")

        splits: tuple[str, str, str] = scan_splits(self.root)

        if self.task == "regression":
            self.train: Dataset = RegressionSubset(self.root + splits[0],
                                                   input_transformations,
                                                   output_transformations,
                                                   input_name,
                                                   output_name,
                                                   )
            self.validation: Dataset = RegressionSubset(self.root + splits[1],
                                                        input_transformations,
                                                        output_transformations,
                                                        input_name,
                                                        output_name,
                                                        )
            self.test: Dataset = RegressionSubset(self.root + splits[2],
                                                  input_transformations,
                                                  output_transformations,
                                                  input_name,
                                                  output_name,
                                                  )
        elif self.task == "classification":
            self.train: Dataset = ClassificationSubset(self.root + splits[0],
                                                       input_transformations,
                                                       )
            self.validation: Dataset = ClassificationSubset(self.root + splits[1],
                                                            input_transformations,
                                                            )
            self.test: Dataset = ClassificationSubset(self.root + splits[2],
                                                      input_transformations,
                                                      )

            # same number of classes in all splits
            if len(self.train.classes) == 0:
                raise ValueError(f"no classes found in {self.train.path}")

            if len(self.train.classes) != len(self.validation.classes):
                raise ValueError(f"number of classes in train and validation splits must be the same")

            if len(self.train.classes) != len(self.test.classes):
                raise ValueError(f"number of classes in train and test splits must be the same")

        else:
            raise ValueError(f"task must be either 'regression' or 'classification'. Found {task}")


class RegressionSubset(Dataset):
    def __init__(self,
                 path: str,
                 input_transformations: list[Transformation],
                 output_transformations: list[Transformation],
                 input_name: str,
                 output_name: str,
                 ) -> None:
        self.root: str = path if path.endswith("/") else path + "/"

        self.input_transformations: list[Transformation] = input_transformations
        self.output_transformations: list[Transformation] = output_transformations
        self.inputs, self.outputs = self._get_data(input_name, output_name)

    def _get_data(self, input_name: str, output_name: str) -> tuple[np.ndarray, np.ndarray]:
        inputs: list[np.ndarray] = []
        outputs: list[np.ndarray] = []

        filenames: list[str] = list(filter(lambda x: x.endswith(".csv"), os.listdir(self.root)))

        if not filenames:
            raise ValueError(f"Directory {self.root} must contain at least one CSV file")

        inputs: list[np.ndarray] = []
        outputs: list[np.ndarray] = []
        for filename in filenames:
            df: pd.DataFrame = pd.read_csv(self.root + filename)

            input = df[input_name].to_numpy()
            output = df[output_name].to_numpy()

            for t in self.input_transformations:
                input = t(input)
            for t in self.output_transformations:
                output = t(output)

            inputs.append(input)
            outputs.append(output)

        return np.concatenate(inputs), np.concatenate(outputs)

    def __len__(self) -> int:
        return self.inputs.shape[0]

    def __getitem__(self, index: int) -> tuple[Tensor, Tensor]:
        return torch.tensor(self.inputs[index]).view(-1), torch.tensor(self.outputs[index]).squeeze()


EXTENSIONS = (".jpg", ".jpeg", ".png", ".ppm", ".bmp", ".pgm", ".tif", ".tiff", ".webp", ".csv")


class Sample(NamedTuple):
    path: str
    label: str


class ClassificationSubset(Dataset):
    def __init__(self,
                 path: str,
                 input_transformations: list[Transformation],
                 ) -> None:
        self.path: str = path if path.endswith("/") else path + "/"

        self.input_transformations: list[Transformation] = input_transformations

        self.classes: list[str] = scan_classes(path)
        self.samples: list[Sample] = []
        for class_name in self.classes:
            class_path: str = os.path.join(path, class_name)

            # continue if a file is found
            if not os.path.isdir(class_path):
                continue

            # get all images paths and corresponding class names
            for root, _, files in os.walk(class_path):
                for file in files:
                    if not file.endswith(EXTENSIONS):
                        continue

                    sample = Sample(os.path.join(root, file), class_name)
                    self.samples.append(sample)

    def __len__(self) -> int:
        return len(self.samples)

    def __getitem__(self, index: int) -> tuple[Tensor, int]:
        sample: Sample = self.samples[index]

        if sample.path.endswith(".csv"):
            sample_array: np.ndarray = pd.read_csv(sample.path).to_numpy()
        else:
            with open(sample.path, "rb") as f:
                img: Image.Image = Image.open(f)
                img = img.convert("RGB")

            sample_array: np.ndarray = np.array(img)

        for transformation in self.input_transformations:
            sample_array = transformation(sample_array)

        label: int = self.classes.index(sample.label)

        return torch.tensor(sample_array), label



import numpy as np

from .transformation import Transformation


class Normalization(Transformation):
    def __init__(self, norm_tech: str) -> None:
        """
        Initialize the normalization technique

        Parameters
        ----------
        norm_tech : str
            Normalization technique to be applied. The following techniques are supported:
            - "arctan"
            - "boxcox"
            - "decimal"
            - "logistic"
            - "mad"
            - "max"
            - "maxabs"
            - "maxmin"
            - "mean"
            - "median"
            - "min"
            - "minabs"
            - "minmax"
            - "power"
            - "quantile"
            - "robust"
            - "softmax"
            - "std"
            - "tanh"
            - "var"
            - "yeojohnson"
            - "zscore"
        """
        self.norm_tech: str = norm_tech

    def __call__(self, x: np.ndarray) -> np.ndarray:
        """
        Apply normalization technique to the input data

        Parameters
        ----------
        x : np.ndarray
            Input data to be normalized

        Returns
        -------
        np.ndarray
            Normalized data
        """
        match self.norm_tech:
            case "arctan":
                return np.arctan(x)
            case "boxcox":
                return np.sign(x) * np.power(np.abs(x), 1 / 2)
            case "decimal":
                return x / 10 ** np.ceil(np.log10(np.max(np.abs(x))))
            case "logistic":
                return 1 / (1 + np.exp(-x))
            case "mad":
                return x / np.median(np.abs(x - np.median(x)))
            case "max":
                return x / np.max(x)
            case "maxabs":
                return x / np.max(np.abs(x))
            case "maxmin":
                return x / (np.max(x) - np.min(x))
            case "mean":
                return x / np.mean(x)
            case "median":
                return x / np.median(x)
            case "min":
                return x / np.min(x)
            case "minabs":
                return x / np.min(np.abs(x))
            case "minmax":
                return (x - np.min(x)) / (np.max(x) - np.min(x))
            case "power":
                return np.sign(x) * np.power(np.abs(x), 1 / 2)
            case "quantile":
                return x / np.quantile(x, 0.95)
            case "robust":
                return x / (np.percentile(x, 75) - np.percentile(x, 25))
            case "softmax":
                return np.exp(x) / np.sum(np.exp(x))
            case "std":
                return x / np.std(x)
            case "tanh":
                return np.tanh(x)
            case "var":
                return x / np.var(x)
            case "yeojohnson":
                return np.sign(x) * np.power(np.abs(x), 1 / 2)
            case "zscore":
                return (x - np.mean(x)) / np.std(x)
            case _:
                raise ValueError("Unknown normalization technique")
